﻿
/*	Copyright Christophe Bertrand.
		This file is available under this license: Ms-PL. See the file License.html in the archive. You will find a copy here: http://www.opensource.org/licenses/ms-pl.html
 */

using System.Threading;
using System.ComponentModel;
using CBdotnet;

namespace AsyncListBox
{
	public class AdvancedAsyncListItemWithPriorityObject : INotifyPropertyChanged
	{
		// This item class contains two texts: the first one has different values, depending on their speed, the second one is always fast.

		// ---------------------------------------------------------------------------------------

		#region Initialization

		// Builder:
		public AdvancedAsyncListItemWithPriorityObject()
		{
			#region Texte 1
			// Here, we declare the functions for Text1:
			{
				// Text1's code is expanded in order to be more readable.

				var valueDelegatesForText1 = new PriorityObjectDelegates<string>();

				// Now we add the delegate values in the order of priority: from the slowest & best to the quickest & worst.
				// This example is simple, but you should manage the ThreadAbortException exception if you need to free (dispose) some resources.

				// First evaluation fonction: slowest & best
				valueDelegatesForText1.Add(
					() =>
					{
						Thread.Sleep(5000); // This simulates a lengthy time before the data being bound to is actualy available.
						return "SloweST text #1";
					});

				// Second evaluation fonction: quicker (but sill slow) and worse
				valueDelegatesForText1.Add(
					() =>
					{
						Thread.Sleep(3000); // This simulates a lengthy time before the data being bound to is actualy available.
						return "Slower text #1";
					});

				Text1 = new PriorityObject<string>(
					"Text1"/*for notifications*/, this.NotifyPropertyChanged, valueDelegatesForText1, true, 
					"Default quickest text"); // a default quick but worst text.
			}
			#endregion

			// - - - - - - - - - - - - - - - - - - - - -- - - - - - - - - - - -- - 

			#region Double1
			// Also, we have a number:
			{
				// Double1's code is reduced in order to be more compact.
				// And Double1 has only one evaluation function; anyway we use PriorityObject so we avoid the complex threads, notifications and dispatchers programming.

				_Double1 = new PriorityObject<double>
					("Double1", this.NotifyPropertyChanged, new PriorityObjectDelegates<double>()
				{
					()=>
					{
						// This simulates a lengthy time before the
						// data being bound to is actualy available.
						Thread.Sleep(2000);
						return 2000.0;
					}
				}, false);
			}
			#endregion

		}

		#endregion

		// ---------------------------------------------------------------------------------------

		#region Properties of this class/item:

		// ------ test async text #1 ----------------

		public PriorityObject<string> Text1 { get; set; }

		// ------ test async double ----------------

		private PriorityObject<double> _Double1;
		// Double1 encapsulates _Double1 in order to avoid the addition of ".Value" in the XAML:
		public double Double1
		{ get { return _Double1.Value; } set { _Double1.Value = value; } }


		// ------------ (quick) Text #2 : --------------

		public string Text2
		{
			get { return "Ordinary fast text #2"; }
		}

		#endregion

		// ---------------------------------------------------------------------------------------

		#region INotifyPropertyChanged
		// --------- INotifyPropertyChanged: ---------------
		public event PropertyChangedEventHandler PropertyChanged;

		// This standard procedure is necessary to PriorityObject.
		private void NotifyPropertyChanged(string info)
		{
			if (PropertyChanged != null)
			{
				PropertyChanged(this, new PropertyChangedEventArgs(info));
			}
		}
		#endregion

		// ---------------------------------------------------------------------------------------
	}
}
