﻿
/*	Copyright Christophe Bertrand.
		This file is available under this license: Ms-PL. See the file License.html in the archive. You will find a copy here: http://www.opensource.org/licenses/ms-pl.html
 */

// Some compiling options you can adjust:

#define CheckIfNull
#define CatchThreadsExceptions
#define AbortUselessThreads // Your evaluator functions must be cancealable at any moment (with ThreadAbortException).
#if SILVERLIGHT
	//#define RequestObjectSDispatcher // For when Silverlight will use several dispatchers.
#endif

// ###################################################################################################################

/* Implementation detail: Silverlight does not accept inter-threads access to PropertyChanged. This complicates the code a bit, and I used define conditions to keep it simple under Dot.net (not silverlight).

 */


using System.Threading;
using System.Collections.Generic;
using System;
using System.Windows.Threading;

namespace CBdotnet
{
	/// <summary>The delegate for the classical NotifyPropertyChanged function.</summary>
	public delegate void NotifyPropertyChangedDelegate(string name);
	/// <summary>The delegate prototyping an evaluation function.</summary>
	public delegate TYPE PriorityObjectDelegate<TYPE>();
	/// <summary>A syntax help, if you are lazy enough to not write <code>List....</code>.   ;)</summary>
	public class PriorityObjectDelegates<TYPE> : List<PriorityObjectDelegate<TYPE>> { }

	/// <summary>
	/// It can be instanciated as a class property.
	/// Its values are evaluated by threads in parallel and have a mutual priority.
	/// </summary>
	public class PriorityObject<TYPE> : IDisposable
	{
		private TYPE _Value;
		private readonly string name;
		private readonly NotifyPropertyChangedDelegate notifyPropertyChangedDelegate;
		private int lastPriorityHavingSetTheValue = int.MaxValue;
		private List<Thread> threads;
		private readonly List<PriorityObjectDelegate<TYPE>> valueDelegates;
		private class threadParam
		{
			public PriorityObjectDelegate<TYPE> Delegate;
			public int ThreadNumber;
		}
		//private bool ThreadsHaveBeenLaunched = false;
#if RequestObjectSDispatcher
		private Dispatcher creatorSDispatcher;
#endif

		/// <summary>
		/// This the main value of this object.
		/// </summary>
		///	<value>Its content varies on the time, depending on the evaluation threads.</value>
		public TYPE Value
		{
			get
			{
				//if ( ! this.ThreadsHaveBeenLaunched)
				if (this.threads == null)
				{
					this.launchThreads();
				}
				return this._Value;
			}
			set
			{
#if CheckIfNull
				// Maybe the _Value is null ? In that case Equals is not adequate.
				if (typeof(TYPE).IsClass)
				{
					if (_Value == null)
					{
						if (value != null)
						{
							this.uncheckedSetValue(value);
						}
						return;
					}
				}
				// Here we know our _Value is not null.
#endif
				if (!this._Value.Equals(value))
				{
					this.uncheckedSetValue(value);
				}
			}
		}
		private void uncheckedSetValue(TYPE value)
		{
			this._Value = value;
#if ! SILVERLIGHT
			this.notifyPropertyChangedDelegate(this.name);
#else
#if RequestObjectSDispatcher
			// Silverlight does not accept inter-threads access to PropertyChanged.
			if (creatorSDispatcher != null)
				this.creatorSDispatcher.BeginInvoke(() => this.notifyPropertyChangedDelegate(this.name));
			else
				this.notifyPropertyChangedDelegate(this.name); // It should provoke a thread exception.
#else
			// If I am not wrong, silverlight 4 do use only one dispatcher.
			System.Windows.Deployment.Current.Dispatcher.BeginInvoke(() => this.notifyPropertyChangedDelegate(this.name));
#endif
#endif
		}

		/// <summary>
		/// Basic Builder.
		/// </summary>
		/// <param name="Name">The name of the property in the class containing that object. For INotifyPropertyChanged.</param>
		/// <param name="notifyPropertyChangedDelegate">The notification function of the class. Usually it is <code>this.NotifyPropertyChanged</code>.</param>
		/// <param name="ValueDelegates">The evaluation functions, in the order of priority: from the slowest and best to the quickest and worst.</param>
		/// <param name="ObjectDispatcher">The Dispatcher of the object/procedure that creates an instance of your class. Only for Silverlight, due to its implementation limitations. Can be null if there is no <code>this.Dispatcher</code> (this can produces exceptions if this procedure is not part of the UI).</param>
		public PriorityObject(
			string Name,
			NotifyPropertyChangedDelegate notifyPropertyChangedDelegate,
			List<PriorityObjectDelegate<TYPE>> ValueDelegates
#if RequestObjectSDispatcher
			, Dispatcher ObjectDispatcher
#endif
)
			: this(Name,notifyPropertyChangedDelegate, ValueDelegates
#if RequestObjectSDispatcher
			, ObjectDispatcher
#endif
, false, default(TYPE))
		{
		}

		/// <summary>
		/// Builder with a Lazy option.
		/// </summary>
		/// <param name="Name">The name of the property in the class containing that object. For INotifyPropertyChanged.</param>
		/// <param name="notifyPropertyChangedDelegate">The notification function of the class. Usually it is <code>this.NotifyPropertyChanged</code>.</param>
		/// <param name="ValueDelegates">The evaluation functions, in the order of priority: from the slowest and best to the quickest and worst.</param>
		/// <param name="ObjectDispatcher">The Dispatcher of the object/procedure that creates an instance of your class. Only for Silverlight, due to its implementation limitations. Can be null if there is no <code>this.Dispatcher</code> (this can produces exceptions if this procedure is not part of the UI).</param>
		/// <param name="Lazy">If true, the evaluation functions are called (and threaded) only if the <c>Value</c> is got. It is similar to <code>Lazy</code>.</param>
		public PriorityObject(
			string Name,
			NotifyPropertyChangedDelegate notifyPropertyChangedDelegate,
			List<PriorityObjectDelegate<TYPE>> ValueDelegates
#if RequestObjectSDispatcher
			, Dispatcher ObjectDispatcher
#endif
, bool Lazy
			)
			: this(Name, notifyPropertyChangedDelegate, ValueDelegates
#if RequestObjectSDispatcher
		, ObjectDispatcher
#endif
, Lazy, default(TYPE))
		{
		}

		/// <summary>
		/// Builder with an initial value.
		/// </summary>
		/// <param name="Name">The name of the property in the class containing that object. For INotifyPropertyChanged.</param>
		/// <param name="notifyPropertyChangedDelegate">The notification function of the class. Usually it is <code>this.NotifyPropertyChanged</code>.</param>
		/// <param name="ValueDelegates">The evaluation functions, in the order of priority: from the slowest and best to the quickest and worst.</param>
		/// <param name="ObjectDispatcher">The Dispatcher of the object/procedure that creates an instance of your class. Only for Silverlight, due to its implementation limitations. Can be null if there is no <code>this.Dispatcher</code> (this can produces exceptions if this procedure is not part of the UI).</param>
		/// <param name="Lazy">If true, the evaluation functions are called (and threaded) only if the <c>Value</c> is got. It is similar to <code>Lazy</code>.</param>
		/// <param name="InitialValue">A default value. It should be very simple and quick to evaluate.</param>
		public PriorityObject(
			string Name,
			NotifyPropertyChangedDelegate notifyPropertyChangedDelegate,
			List<PriorityObjectDelegate<TYPE>> ValueDelegates
#if RequestObjectSDispatcher
			, Dispatcher ObjectDispatcher
#endif
, bool Lazy,
			TYPE InitialValue)
		{
			this._Value = InitialValue;
			this.name = Name;
			this.notifyPropertyChangedDelegate = notifyPropertyChangedDelegate;
			this.valueDelegates = ValueDelegates;
#if RequestObjectSDispatcher
			if (ObjectDispatcher != null)
				this.creatorSDispatcher = ObjectDispatcher;
			else
			{
				this.creatorSDispatcher = System.Windows.Deployment.Current.Dispatcher;
				if (this.creatorSDispatcher == null)
					this.creatorSDispatcher = System.Windows.Application.Current.MainWindow.Dispatcher;
			}
#endif

			if (!Lazy)
				this.launchThreadsWithoutLock();
		}

		private void launchThreads()
		{
			lock (this.valueDelegates) // We could have create a new object to lock, but this one already exists.  ;)
			{
				if (this.threads == null) // Tested twice due to thread parallelism (here there's lock).
					launchThreadsWithoutLock();
			}
		}
		private void launchThreadsWithoutLock()
		{
			//this.ThreadsHaveBeenLaunched = true;

			this.threads = new List<Thread>();
			foreach (PriorityObjectDelegate<TYPE> current in this.valueDelegates)
			{
				this.threads.Add(new Thread(new ParameterizedThreadStart(this.ThreadProcedure)));
			}
			for (int i = 0; i < this.threads.Count; i++)
			{
				PriorityObject<TYPE>.threadParam threadParam = new PriorityObject<TYPE>.threadParam();
				threadParam.ThreadNumber = i;
				threadParam.Delegate = this.valueDelegates[i];
#if DEBUG
				this.threads[i].Name = "PriorityObject{"+this.name+":"+i+"}";
#endif
#if ! SILVERLIGHT
				this.threads[i].Priority = ThreadPriority.BelowNormal;// Lower priority than the UI.
#endif
				this.threads[i].Start(threadParam);
			}
		}

		// The thread procedure will call the user's value procedure:
		private void ThreadProcedure(object o)
		{
			try
			{
				PriorityObject<TYPE>.threadParam threadParam = o as PriorityObject<TYPE>.threadParam;
				if (threadParam != null)
				{
					if (threadParam.Delegate != null)
					{
						this.internalSetValue(threadParam.Delegate(), threadParam.ThreadNumber);
					}
				}
			}
			catch (ThreadAbortException) { } // This particular exception is normal in this context.
#if CatchThreadsExceptions
#if ! DEBUG
			catch {} // Avoid to close the entire application.
#else // accept exceptions to facilitate debugging
			catch (Exception e)
			{
				if (System.Diagnostics.Debugger.IsAttached) // If we are debugging in Visual Studio (or other IDE).
					throw e;
				// Otherwize, we choose to not close the application.
			}
#endif // DEBUG
#endif // CatchThreadsExceptions
		}

		private void internalSetValue(TYPE value, int ThreadNumber)
		{
			lock (this.valueDelegates) // We could have create a new object to lock, but this one already exists.  ;)
			{
				if (ThreadNumber < this.lastPriorityHavingSetTheValue)
				{
					this.lastPriorityHavingSetTheValue = ThreadNumber;

					/*if (typeof(TYPE).IsClass)
					{
						// Spécial test: if a higher-priority thread wants to replace a not-null Value by null, we refuse.

						if ((this._Value != null) && ....
						{
					 		......
						}
					}
					else*/
						this.Value = value;

#if AbortUselessThreads
						// Now we can abort the lower-priority threads:
					for (int i = ThreadNumber + 1; i < this.threads.Count; i++)
					{
						if (this.threads[i] != null)
							if (this.threads[i].IsAlive)
							{
								this.threads[i].Abort();
								this.threads[i] = null;
							}
					}
#endif
				}
			}
		}

		// --------------------

		/// <summary>
		/// Dispose from IDisposable. Tries to abord the evaluation threads (if compiled with AbortUselessThreads defined).
		/// </summary>
		public void Dispose()
		{
			lock (this.valueDelegates) // We could have create a new object to lock, but this one already exists.  ;)
			{
				this.lastPriorityHavingSetTheValue = -1; // To prevent threads from changing the Value.
#if AbortUselessThreads
				if (threads != null)
				{
					// We abort all our threads:
					for (int i = 0; i < this.threads.Count; i++)
					{
						if (this.threads[i] != null)
							if (this.threads[i].IsAlive)
							{
								this.threads[i].Abort();
								this.threads[i] = null;
							}
					}
				}
#endif
			}
		}
	}
}
